
#####################################################################
#
# Application Generator; version 2.0
# (C) 2003-2005, Alex S. <alex@theparticle.com>
# http://www.theparticle.com/
#
#####################################################################

use pxml;

use strict;

my $generator = "Application Generator; Version 2.0; ".
                "(C) 2003-2005, Alex S.";

my $dbgen = "gen/dbgen.sql";

my $tab = ' 'x4;

#####################################################################
## load all types.
#####################################################################
my %gen_types;
my $metatypes = new pxml(file => "metatypes.xml");
for my $type ($metatypes->query('*/@type')){
    $gen_types{ $type->query('%name') } = {
        sql => $type->query('sql/$'),
        str => $type->query('@str') ? 1:0,
    };
}

#####################################################################
## load all objects.
#####################################################################
my @objects;
my %objects;
my $metadata = new pxml(file => "metadata.xml");
for my $object ($metadata->query('*/@object')){
    push @objects,make_object($object);
    $objects{ $objects[-1]->{name} } = $objects[-1];
}

#####################################################################
## Call methods to generate code.
#####################################################################

make_sql(\@objects);
make_value_objects(\@objects);

make_listview_form(\@objects);
make_detailview_form(\@objects);

make_createview_form(\@objects);
make_updateview_form(\@objects);
make_createaction_form(\@objects);
make_updateaction_form(\@objects);

#make_choosermenuview_form(\@objects);

#####################################################################
## make create action.
#####################################################################
sub make_updateaction_form {
    my ($objects) = @_;
    for my $object (@{$objects}){    
        open my $out,'>gen/'.$object->{name}.'UpdateAction.php' or die $!;

print $out qq|<?php
// $generator
include ("$object->{name}.php");

\$form = array();
|.join("\n",map {
    qq{\$form['$_->{name}'] = trim(\$HTTP_POST_VARS['$_->{name}']);}
} @{$object->{attrib}}).qq|

\$alliswell = 1;
if(1==0){
|.join("\n",map {
    qq|}else if(!\$form['$_->{name}']){
    \$alliswell = 0;
    \$HTTP_SESSION_VARS['error'] = "$_->{fname} needs to be specified.";|
} grep { $_->{pkey} } @{$object->{attrib}}).qq|
|.join("\n",map {
    qq|}else if(!\$form['$_->{name}']){
    \$alliswell = 0;
    \$HTTP_SESSION_VARS['error'] = "$_->{fname} needs to be specified.";|
} grep { $_->{required} } @{$object->{attrib}}).qq|
}else{
    // create the object
    \$obj = new $object->{name}();
|.join("\n",map {qq{\$obj->$_->{name} = \$form['$_->{name}'];}} @{$object->{attrib}}).qq|

    dbUpdate$object->{name}(\$obj);
    if(db_duplicate()){
        \$alliswell = 0;
        \$HTTP_SESSION_VARS['error'] = 'Duplicate value detected. Pick a different name.';
|.join("\n",map {qq|unset(\$form['$_->{name}']);|} grep { $_->{unique} } @{$object->{attrib}}).qq|
    }
    if(\$alliswell){
        Header("Location: get.php?q=some/where/blah&|.
join('&',map {qq|$_->{name}=".\$form['$_->{name}']."|} grep { $_->{pkey} } @{$object->{attrib}}).qq|");
        exit;
    }
}
\$HTTP_SESSION_VARS['form'] = serialize(\$form);
Header("Location: get.php?q=blah/blah/update&|.
join('&',map {qq|$_->{name}=".\$form['$_->{name}']."|} grep { $_->{pkey} } @{$object->{attrib}}).qq|");
exit;
?>|;
        close $out;
    }
}



#####################################################################
## make create action.
#####################################################################
sub make_createaction_form {
    my ($objects) = @_;
    for my $object (@{$objects}){    
        open my $out,'>gen/'.$object->{name}.'CreateAction.php' or die $!;

print $out qq|<?php
// $generator
include ("$object->{name}.php");

\$form = array();
|.join("\n",map {
    qq{\$form['$_->{name}'] = trim(\$HTTP_POST_VARS['$_->{name}']);}
} @{$object->{attrib}}).qq|

\$alliswell = 1;
if(1==0){
|.join("\n",map {
    qq|}else if(!\$form['$_->{name}']){
    \$alliswell = 0;
    \$HTTP_SESSION_VARS['error'] = "$_->{fname} needs to be specified.";|
} grep { $_->{required} } @{$object->{attrib}}).qq|
}else{
    // create the object
    \$obj = new $object->{name}();
|.join("\n",map {qq{\$obj->$_->{name} = \$form['$_->{name}'];}} @{$object->{attrib}}).qq|
    \$r = dbCreate$object->{name}(\$obj);
    if(!\$r){
        \$alliswell = 0;
        if(db_duplicate()){
	    \$HTTP_SESSION_VARS['error'] = 'Duplicate value detected. Pick a different name.';
|.join("\n",map {qq|unset(\$form['$_->{name}']);|} grep { $_->{unique} } @{$object->{attrib}}).qq|
        }else{
            \$HTTP_SESSION_VARS['error'] = 'Database Error. Please try again later.';
        }
    }

    if(\$alliswell){
        Header("Location: get.php?q=list/of/these");
        exit;
    }
}
\$HTTP_SESSION_VARS['form'] = serialize(\$form);
Header("Location: get.php?q=crate/objthing");
exit;
?>|;
        close $out;
    }
}



#####################################################################
## make update view form.
#####################################################################
sub make_updateview_form {
    my ($objects) = @_;
    for my $object (@{$objects}){    
        open my $out,'>gen/'.$object->{name}.'UpdateView.php' or die $!;

print $out qq|<h2>Update $object->{name}!</h2>
<?php

// $generator

include ("$object->{name}.php");

\$form = \$HTTP_SESSION_VARS['form'];
unset(\$HTTP_SESSION_VARS['form']);

if(\$form){
    \$formdata = unserialize(\$form);
}else{
    \$formdata = array();
}
if(\$HTTP_SESSION_VARS['error']){
    ?><h3 class="formerror"><?php    
echo(\$HTTP_SESSION_VARS['error']);
unset(\$HTTP_SESSION_VARS['error']);
    ?></h3><?php
}
?>
<SCRIPT LANGUAGE="JAVASCRIPT">|.q#
    function isAllSpaces(s){
        for(var i = 0;i<s.length;i++){
            if(s.charAt(i) != ' ')
                return false;
        }
        return true;
    }
    function isEmail(s){
        if(isAllSpaces(s))
            return false;
        if(s.indexOf('@') <= 0 || s.indexOf('@') == s.length - 1)
            return false;
        return true;
    }
    function validate(testform){#.
    join("\n",map {
       qq(\n        if (isAllSpaces(testform.$_->{name}.value)) {
            window.alert("$_->{fname} cannot be left blank.");
            return false;
       })
    } grep { $_->{required} } @{$object->{attrib}})
       .qq|\n        return true;
    }
</SCRIPT>
<?php\n|;
for my $r (grep { $_->{pkey} } @{$object->{attrib}}){
    print $out '$'.$r->{name}." = ".($r->{str} ? "":"(int)").
        "trim(\$HTTP_GET_VARS['".$r->{name}."']);\n";
}
print $out "\$rec = dbGet$object->{name} (".
    join(',',map {'$'.$_->{name}} grep { $_->{pkey} } @{$object->{attrib}}).");\n";

print $out qq|
if(!\$rec){
    ?><h3 class="formerror">Record Not Found (|.  
    join(',',map {$_->{name}.'=<?=$'.$_->{name}.'?>'} grep { $_->{pkey} } @{$object->{attrib}})
    .qq|)</h3><?php
}else{
?>
<p>
<form action="do.php?q=blah/action" method="POST" onsubmit="return validate(this);">
<table class="formtable">|.
    join("\n",map {
       qq(
    <tr class="formtr">
        <td class="formtdname"><nobr>$_->{fname}:</nobr></td>
        <td class="formtdvalue">
            <?=\$rec->$_->{name}?>
            <input type="hidden" name="$_->{name}" value="<?=\$$_->{name}?>">
        </td>
    </tr>)
    } grep { $_->{auto} } @{$object->{attrib}}).    
    join("\n",map {
       qq(
    <tr class="formtr">
        <td class="formtdname"><nobr>$_->{fname}:</nobr></td>
        <td class="formtdvalue">
            <input type="text" name="$_->{name}" value="<? echo(\$formdata['$_->{name}']? \$formdata['$_->{name}']:\$rec->$_->{name})?>" class="formtextfield">
        </td>
    </tr>)
    } grep { !$_->{auto} } @{$object->{attrib}}).
  qq|<tr class="formtr">
    <td colspan="2" class="formtdsubmit">
        <input type="submit" value="Update $object->{name}">
    </td>
  </tr>
  </table>
</p>
<?php
}
unset(\$form,\$formdata);
?>|;

        close $out;
    }
}



#####################################################################
## make create view form.
#####################################################################
sub make_createview_form {
    my ($objects) = @_;
    for my $object (@{$objects}){    
        open my $out,'>gen/'.$object->{name}.'CreateView.php' or die $!;

print $out qq|<h2>Create $object->{name}!</h2>
<?php

// $generator

\$form = \$HTTP_SESSION_VARS['form'];
unset(\$HTTP_SESSION_VARS['form']);

if(\$form){
    \$formdata = unserialize(\$form);
}else{
    \$formdata = array();
}
if(\$HTTP_SESSION_VARS['error']){
    ?><h3 class="formerror"><?php    
echo(\$HTTP_SESSION_VARS['error']);
unset(\$HTTP_SESSION_VARS['error']);
    ?></h3><?php
}
?>
<SCRIPT LANGUAGE="JAVASCRIPT">|.q#
    function isAllSpaces(s){
        for(var i = 0;i<s.length;i++){
            if(s.charAt(i) != ' ')
                return false;
        }
        return true;
    }
    function isEmail(s){
        if(isAllSpaces(s))
            return false;
        if(s.indexOf('@') <= 0 || s.indexOf('@') == s.length - 1)
            return false;
        return true;
    }
    function validate(testform){#.
    join("\n",map {
       qq(\n        if (isAllSpaces(testform.$_->{name}.value)) {
            window.alert("$_->{fname} cannot be left blank.");
            return false;
       })
    } grep { $_->{required} } @{$object->{attrib}})
       .qq|\n        return true;
    }
</SCRIPT>
<p>
<form action="do.php?q=blah/blah/action" method="POST" onsubmit="return validate(this);">
<table class="formtable">|.
    join("\n",map {
       qq(\n<tr class="formtr">
    <td class="formtdname"><nobr>$_->{fname}:</nobr></td>
    <td class="formtdvalue"><input type="text" 
name="$_->{name}" value="<?=\$formdata['$_->{name}']?>" class="formtextfield"></td>
  </tr>)
    } grep { !$_->{auto} } @{$object->{attrib}})
  .qq|\n<tr class="formtr">
    <td colspan="2" class="formtdsubmit">
     <input type="submit" value="Create $object->{name}">
    </td>
  </tr>
  </table>
</form>
</p>
<?php
    unset(\$form,\$formdata);
?>|;
        close $out;
    }
}




#####################################################################
## make detailview form.
#####################################################################
sub make_detailview_form {
    my ($objects) = @_;
    for my $object (@{$objects}){    
        open my $out,'>gen/'.$object->{name}.'DetailView.php' or die $!;
print $out qq|<?php
// $generator

include ("$object->{name}.php");\n|;

for my $r (grep { $_->{pkey} } @{$object->{attrib}}){
    print $out '$'.$r->{name}." = ".($r->{str} ? "":"(int)").
        "trim(\$HTTP_GET_VARS['".$r->{name}."']);\n";
}
print $out "\$rec = dbGet$object->{name} (".
    join(',',map {'$'.$_->{name}} grep { $_->{pkey} } @{$object->{attrib}}).");\n";

print $out "?><h2>View $object->{name} Details</h2><?php\n";

print $out qq|
if(!\$rec){
    ?><h3 class="formerror">Record Not Found (|.  
    join(',',map {$_->{name}.'=<?=$'.$_->{name}.'?>'} grep { $_->{pkey} } @{$object->{attrib}})
    .qq|)</h3><?php
}else{
?>
<p>
<table class="formtable">|.
  join("\n",map {
  qq(\n<tr class="formtr">
    <td class="formtdname"><nobr>$_->{fname}:</nobr></td>
    <td class="formtdvalue">
        <?=\$rec->$_->{name}?>
    </td>
  </tr>)
  } @{$object->{attrib}})
  .qq|<tr class="formtr">
    <td colspan="2" class="formtdsubmit">
        <A href="get.php?q=whatever/whatever&|.
         join('&',map {$_->{name}.'=<?=$'.$_->{name}."?>"} 
            grep { $_->{pkey} } @{$object->{attrib}})         
         .qq|">Update $object->{name}</A>
    </td>
  </tr>
  </table>
</p>
<?php
}
?>\n|;
        close $out;
    }
}

#####################################################################
## make listview form.
#####################################################################
sub make_listview_form {
    my ($objects) = @_;
    for my $object (@{$objects}){    
        open my $out,'>gen/'.$object->{name}.'ListView.php' or die $!;
print $out qq|<?php
// $generator

include ("$object->{name}.php");

\$form = array();
\$form['page'] = (int)trim(\$HTTP_GET_VARS['page']);
\$recs_per_page = 3;

\$start = \$form['page'] * \$recs_per_page;

\$recs = dbGetAll$object->{pname}(\$start,\$recs_per_page+1);

if(sizeof(\$recs)){
        ?><TABLE>
                <TR class="tabletrheader">
|;

        print $out 
         join("\n",map {"<td>".$_->{fname}."</td>"} @{$object->{attrib}});

print $out qq|
                </TR>
        <?php
        for (\$i=0;\$i<sizeof(\$recs) && \$i < \$recs_per_page;\$i++){
            \$value = \$recs[\$i];
            ?><TR>
|;
                
            print $out 
             join("\n",map {"<td><?=\$value->".$_->{name}."?></td>"} @{$object->{attrib}});
                
print $out qq|
              </TR>
            <? 
        }
        ?>
        </TABLE>

    <table border="0">
        <tr>
        <TD WIDTH="50%" ALIGN="left"><?php
            if(\$form['page']){
                \$page = \$form['page']-1;
                ?><A HREF="test.php?page=<?=\$page?>">&lt;&lt;Previous Page</A><?php
            }
        ?></TD>
        <TD WIDTH="50%" ALIGN="right"><?php
            if(sizeof(\$recs) > \$recs_per_page){
                \$page = \$form['page']+1;
                ?><A HREF="test.php?page=<?=\$page?>">Next Page &gt;&gt;</A><?php
            }
        ?></TD>
        </tr>
    </table><?php
}else{
    ?>No records found.<?php
}

?>
|;              
        close $out;
    }
}


#####################################################################
## make sql schema file.
#####################################################################
sub make_value_objects {
    my ($objects) = @_;

    for my $object (@{$objects}){
    
        open my $out,'>gen/'.$object->{name}.'.php' or die $!;
        
        print $out qq{<?php 
// $generator
// $object->{name}.php

// THIS FILE IS GENERATED FROM metadata.xml.
// DO NOT UPDATE THIS FILE.
// ANY CHANGES TO THIS FILE WILL BE LOST!!!
// ALL CHANGES MUST BE MADE TO metadata.xml.
\n};

        print $out qq|
// $object->{name}
// $object->{description}

class $object->{name} {
|;
        for my $attrib (@{$object->{attrib}}){
            print $out $tab.'var $'.$attrib->{name}.";\n";
        }
      
        print $out $tab."function $object->{name} (\$arr){\n";
        for my $attrib (@{$object->{attrib}}){
            print $out $tab.$tab.'$this->'.$attrib->{name}.
                '=$arr[\''.$attrib->{sqlname}.'\'];'."\n";
        }
        print $out "    }\n";
        print $out "}\n";


        #############################################################
        # create object query. (object)
        print $out qq|
// create $object->{name} query. accepts the object, and returns the newly created id.
function dbCreate$object->{name}(\$o) {
    \$ret = NULL;
    \$sql = "INSERT INTO $object->{sqlname} (";\n|;
        print $out $tab.qq|\$sql .= "|.
            join(',',map {$_->{sqlname}} grep { !$_->{auto} } @{$object->{attrib}}).
            qq|";\n|;
        print $out $tab.qq|\$sql .= ") VALUES (";\n|;
        print $out join(qq|\n$tab\$sql .= ",";\n|,
            map {
                $tab.qq|\$sql .= isset(\$o->$_->{name}) ? "'".\$o->$_->{name}."'":"NULL";|
            } grep { !$_->{auto} } @{$object->{attrib}});
        print $out qq|
    \$sql .= ");";
    if(\$result = db_query(\$sql)){
        \$ret = db_insertid();
    }
    return \$ret;
}\n|;


        #############################################################
        # update object query. (object with valid id)        
        print $out qq|
// update $object->{name} query. accepts the object, and updates it.
function dbUpdate$object->{name}(\$o) {
    \$ret = NULL;
    \$sql = "UPDATE $object->{sqlname} SET ";\n|;
        print $out join(qq|\n$tab\$sql .= ",";\n|,
            map {
                $tab.qq|\$sql .= "$_->{sqlname}=".(isset(\$o->$_->{name}) ? "'".\$o->$_->{name}."'":"NULL");|
            } grep { !$_->{pkey} } @{$object->{attrib}});
        print $out "\n".$tab.qq|\$sql .= " WHERE ";\n|;
        print $out join(qq|\n$tab\$sql .= " AND ";\n|,
            map {
                $tab.qq|\$sql .= "$_->{sqlname}='".\$o->$_->{name}."'";|
            } grep { $_->{pkey} } @{$object->{attrib}});
        print $out qq|
    \$sql .= ";";
    \$ret = db_query(\$sql);
    return \$ret;
}\n|;

        #############################################################
        # delete object query. (id)
        print $out qq|
// delete $object->{name} query. accepts the primey key.
function dbDelete$object->{name}(|;
        print $out join(',',
            map {'$'.$_->{name}} grep { $_->{pkey} } @{$object->{attrib}}
        );
        print $out qq|) {
    \$ret = NULL;
    \$sql = "DELETE $object->{sqlname} WHERE ";\n|;
        print $out join(qq|\n$tab\$sql .= " AND ";\n|,
            map {
                $tab.qq|\$sql .= "$_->{sqlname}='".\$$_->{name}."'";|
            } grep { $_->{pkey} } @{$object->{attrib}});
        print $out qq|
    \$sql .= ";";
    \$ret = db_query(\$sql);
    return \$ret;
}\n|;

        #############################################################
        # get object query. (id)
        print $out qq|
// get $object->{name} query. accepts the primey key.
function dbGet$object->{name}(|;
        print $out join(',',
            map {'$'.$_->{name}} grep { $_->{pkey} } @{$object->{attrib}}
        );
        print $out qq|) {
    \$ret = NULL;
    \$sql = "SELECT * FROM $object->{sqlname} WHERE ";\n|;
        print $out join(qq|\n$tab\$sql .= " AND ";\n|,
            map {
                $tab.qq|\$sql .= "$_->{sqlname}='".\$$_->{name}."'";|
            } grep { $_->{pkey} } @{$object->{attrib}});
        print $out qq|
    \$sql .= ";";
    if(\$result = db_query(\$sql)){
        if (\$row = db_fetch(\$result)){
            \$ret = new $object->{name}(\$row);
        }
        db_freeres(\$result);
    }
    return \$ret;
}\n|;

        #############################################################
        # get all objects query. 
        print $out qq|
// get $object->{name} query. accepts the primey key.
function dbGetAll$object->{pname}(\$start = 0,\$end = 4096) {
    \$ret = NULL;
    \$sql = "SELECT * FROM $object->{sqlname} LIMIT ".\$start.",".\$end.";";
    if(\$result = db_query(\$sql)){
        \$ret = array();
        while (\$row = db_fetch(\$result)){
            \$ret[] = new $object->{name}(\$row);
        }
        db_freeres(\$result);
    }
    return \$ret;
}\n|;

        #############################################################
        # get count of objects.
        print $out qq|
// get $object->{name} query. accepts the primey key.
function dbCount$object->{pname}() {
    \$ret = NULL;
    \$sql = "SELECT COUNT(*) AS N FROM $object->{sqlname};";
    if(\$result = db_query(\$sql)){
        if (\$row = db_fetch(\$result)){
            \$ret = \$row['N'];
        }
        db_freeres(\$result);
    }
    return \$ret;
}\n|;


        #############################################################
        # work on the queries
        foreach my $query (@{$object->{query}}){
            
            my $params = join(',',map { '$'.$_->{name} } @{$query->{param}});
            
            my $sql = $query->{sql};
        
            for my $param (@{$query->{param}}){
                $sql =~ s|\$$param->{name}\$|\$$param->{name}|gs;
            }
            
            print $out qq|
//$query->{description}
function query$query->{name}($params) {
    \$ret = NULL;
    \$sql = "$sql";
    if(\$result = db_query(\$sql)){
|;
            if($query->{single}){
                print $out "        if (\$row = db_fetch(\$result)){\n";
                print $out "            \$ret = \$row;\n" if $query->{raw};
                print $out "            \$ret = new $object->{name}(\$row);\n" if !$query->{raw};
                print $out "        }\n";
                print $out "        db_freeres(\$result);\n";
            }elsif($query->{multiple}){
                print $out "        \$ret = array();\n";
                print $out "        while (\$row = db_fetch(\$result)){\n";
                print $out "            \$ret[] = new $object->{name}(\$row);\n" if !$query->{raw};
                print $out "            \$ret[] = \$row;\n" if $query->{raw};
                print $out "        }\n";
                print $out "        db_freeres(\$result);\n";
            }
            if($query->{autoinc}){
                print $out "        \$ret = db_insertid();\n";
            }
            print $out "    }\n";
            print $out "    return \$ret;\n";
            print $out "}\n";
        
        }
       
        print $out "\n?>";
        close $out;
    }
}




#####################################################################
## make sql schema file.
#####################################################################
sub make_sql {
    my ($objects) = @_;
    
    open my $out,">$dbgen" or die $!;
    binmode $out;

    print $out qq|-- $generator\n\n|;
    print $out qq|-- This file is GENERATED from metadata.xml\n|;
    print $out qq|-- DO NOT MODIFY THIS FILE!!!\n|;
    print $out qq|-- All changes should be made to metadata.xml instead.\n\n|;

    for my $object (@{$objects}){
        print $out qq|-- $object->{description}\n|;
        print $out qq|DROP TABLE $object->{sqlname};\n\n|;
        print $out qq|CREATE TABLE $object->{sqlname} (\n|;
        for my $attrib (@{$object->{attrib}}){
            print $out $tab.$attrib->{sqlname}." ".$attrib->{sqltype};
            print $out " DEFAULT $attrib->{sqldefault}" if $attrib->{sqldefault};
            print $out " NOT NULL" if $attrib->{required};
            print $out " AUTO_INCREMENT" if $attrib->{auto};
            print $out ",";
            print $out $tab."-- $attrib->{description}" if $attrib->{description};
            print $out "\n";
        }
        print $out $tab."PRIMARY KEY(".$object->{sqlpkey}.")";
        for my $key (@{$object->{key}}){
            my $attrib = $object->{attrib_names}{$key};
            print $out ",\n".$tab."KEY(".$attrib->{sqlname}.")";
        }
        for my $unique (@{$object->{unique}}){
            my $attrib = $object->{attrib_names}{$unique};
            print $out ",\n".$tab."UNIQUE(".$attrib->{sqlname}.")";
        }
        print $out "\n);\n\n";
    }
    close $out;
}

#####################################################################
## convert my type into an SQL type
#####################################################################
sub to_SQL_type {
    my ($type,$size) = @_;
    my $v = $gen_types{$type}{sql};
    $v .= "($size)" if $size > 0;
    return $v;
}

#####################################################################
## given an XML object root, build a hash with all `important' data
#####################################################################
sub make_object {

    my ($object) = @_;

    my $obj = {};

    $obj->{name} = $object->query('%name');

    $obj->{pname} = $object->query('pname/$');
    $obj->{pname} = $obj->{name}.'s' if !$obj->{pname};
    
    $obj->{description} = $object->query('description/$');
    $obj->{description} = $obj->{name} if !$obj->{description};

    # print "$obj->{name}: $obj->{description}\n";

    $obj->{sqlname} = lc($obj->{name});

    $obj->{pkey} = [];
    $obj->{key} = [];
    $obj->{unique} = [];

    $obj->{attrib} = [];
    $obj->{attrib_names} = {};

    my @sqlpkey = ();

    for my $attrib ($object->query('@attribute')){
        
        my $attr = {};
        $attr->{name} = $attrib->query('%name');
        $attr->{type} = $attrib->query('type/$');
        $attr->{size} = $attrib->query('size/$');
        $attr->{sqldefault} = $attrib->query('default/$');
        $attr->{str} = $gen_types{ $attr->{type} }{str};
        $attr->{sqlname} = lc($attr->{name});
        $attr->{sqltype} = to_SQL_type($attr->{type},$attr->{size});
        $attr->{required} = 1 if $attrib->query('@required');
        $attr->{auto} = 1 if $attrib->query('@auto');
        $attr->{pkey} = 1 if $attrib->query('@pkey');
        $attr->{key} = 1 if $attrib->query('@key') || $attrib->query('@index');
        $attr->{unique} = 1 if $attrib->query('@unique');
        $attr->{password} = 1 if $attrib->query('@password');

        # field name
        $attr->{fname} = $attrib->query('fname/$');
        $attr->{fname} = $attr->{name} if !$attr->{fname};

        $attr->{description} = $attrib->query('description/$');
        $attr->{description} = $attr->{fname} if !$attr->{description};

        push @{$obj->{pkey}},$attr->{name} if $attr->{pkey};
        push @{$obj->{auto}},$attr->{name} if $attr->{auto};
        push @{$obj->{key}},$attr->{name} if $attr->{key};
        push @{$obj->{unique}},$attr->{name} if $attr->{unique};
        push @sqlpkey,$attr->{sqlname} if $attr->{pkey};
        push @{$obj->{attrib}},$attr;
        ${$obj->{attrib_names}}{$attr->{name}} = $attr;
    }
    $obj->{sqlpkey} = join ',',@sqlpkey;
    $obj->{firstkey} = ${$obj->{pkey}}[0];
    $obj->{firstauto} = ${$obj->{auto}}[0];

#    <query name="PersonByName">
#        <description>Given a first name and last name, finds people that match'em.</description>
#        <parameter name="lname"/>
#        <parameter name="fname"/>
#        <sql>SELECT p.* FROM [Person] p WHERE p.[lname] = $lname$ AND p.[fname] = $fname$</sql>
#        <single/>
#    </query>

    $obj->{query} = [];
    
    for my $query ($object->query('@query')){        
        my $attr = {};

        $attr->{name} = $query->query('%name');
        $attr->{description} = $query->query('description/$');
        $attr->{raw} = 1 if $query->query('@raw');
        $attr->{single} = 1 if $query->query('@single');
        $attr->{multiple} = 1 if $query->query('@multiple');
        $attr->{autoinc} = 1 if $query->query('@autoinc');
        $attr->{sql} = $query->query('sql/$');

        $attr->{param} = [];
        for my $param ($query->query('@parameter')){
            my $parm = {};
            $parm->{name} = $param->query('%name');
            push @{$attr->{param}},$parm;
        }
        push @{$obj->{query}},$attr;
    }

    return $obj;
}


