#####################################################################
##
## Particle's simple XML Module, Version 1.0
##     (not designed for speed)
## (C) 2005, Alex S. <alex@theparticle.com>
## http://www.theparticle.com/
##
#####################################################################

package pxml;

use strict;

use overload (
    'fallback'  => 1,
    '""'  => 'as_string',
);

######################################################################
## create a new instance of xml object.
######################################################################
sub new {
    my ($class, %args) = @_;
    my $this = { %args };
    if($args{file}){
        my $root = parse(loadFile($this->{file}));
        for my $key (keys %{$root}){
            $this->{$key} = $root->{$key};
        }
    }
    return bless $this, $class;
}

######################################################################
## output the entire parsed xml.
######################################################################
sub as_string {
    my ($this) = @_;
    my @str;
    my @stack;
    push @stack,[$this,0];
    while($#stack >= 0){
        my ($tag,$tab) = @{pop @stack};
        if($tag->{type} eq 'tag'){
            push @str,(' 'x $tab)."<".$tag->{value};
            for(sort keys %{$tag->{attrib}}){
                push @str,' '.$_.'="'.$tag->{attrib}{$_}.'"';
            }
            push @str,">\n";
            push @stack,[ { type=>'ctag', value=>$tag->{value} },$tab];
            for( reverse @{ $tag->{children} } ){
                push @stack,[$_,$tab+3];
            }
        }elsif($tag->{type} eq 'ctag'){
            push @str,(' 'x $tab).'</'.$tag->{value}.">\n";
        }elsif($tag->{type} eq 'text'){
            push @str,(' 'x $tab);
            push @str,($tag->{value} =~ m/>|</ ? 
                    '<![CDATA['.$tag->{value}.']]>':$tag->{value});
            push @str,"\n";
        }
    }
    return join '',@str;
}

######################################################################
## returns contents of file
######################################################################
sub loadFile {
   my ($name) = @_;
   open my $in,$name or die qq{$!: "$name"};
#   binmode $in;
   read $in,my $cont,-s $name or die qq{$!: "$name"};
   close $in;
   return $cont;
}

######################################################################
## XML Parser & Query Code
######################################################################


######################################################################
# xml parser; returns root node
######################################################################
sub parse {
    $_ = shift;
    my @strings;
    s{<!\[CDATA\[(.*?)\]\]>}{push @strings,$1;'<#'.$#strings.'/>'}xsge;
    s|<!--.*?-->||gs;
    push my @stack, { type=>'root', value=>'~', children=>[]};
    while(m|(.*?)<(.+?)>|gs){
        for(my $t = $1){
            s/^\s+|\s+$//gs; s/\s+/ /gs; s|\&lt;|<|gs; s|\&gt;|>|gs;
            s|\&amp;|\&|gs; s|\&apos;|\'|gs; s|\&quot;|\"|gs;
            push @{ $stack[$#stack]->{children} },
                new pxml(type=>'text',value=>$_) if length;
        }
        my $t = $2;
        if($t !~ m|/|){ # opening tag
            my ($tag,$attr) = split /\s+/,$t,2;
            push @stack, new pxml(type=>'tag', value=>$tag, attrib=>{}, children=>[]);
            $stack[$#stack]->{attrib}{$1} = $2 while($attr =~ m|\s*(.*?)="(.*?)"|g);
        }elsif($t =~ m|^/|){ # closing tag
            my $s = pop @stack;
            push @{ $stack[$#stack]->{children} },$s if $t eq "/$s->{value}";
        }elsif($t =~ m|#(\d+)|){    # CDATA section
            push @{ $stack[$#stack]->{children} },
                new pxml(type=>'text',value=>$strings[$1]) if $strings[$1];
        }elsif($t =~ m|/$|){
            $t =~ s|/$||;
            my ($tag,$attr) = split /\s+/,$t,2;
            my $o = new pxml(type=>'tag',value=>$tag, attrib=>{}, children=>[]);
            $o->{attrib}{$1} = $2 while($attr =~ m|\s*(.*?)="(.*?)"|g);
            push @{ $stack[$#stack]->{children} },$o;
        }
    }
    return $stack[0]{children}[0];
}

######################################################################
# ---queries---
#  schema/object/%     (retrieves all attributes)
#  schema/object/%name (retrieves name attribute for each object)
#  schema/object/@      (retrieves all children)
#  schema/@object       (retrieves all object children)
#  schema/object/$     (retrieves any text in each object)
#  /schema/@object      (ensures that we start search from root)
#  schema/*/@     (* indicates any tag)
# query XML for various tags/attributes
#
# NOTE: this method is the workhorse of this whole package; but
#       it is very buggy, and not very clean (this method can
#       and will break just when you need it!)
#
# TODO: fix, and make this code cleaner.
#
######################################################################
sub query {
    my ($this,$query) = @_;
    my @query = split '/',$query;

    my @result;

    # prepare query
    $query[0] = '~' if !$query[0];
    my $result = pop @query;       # what are we looking for?
    $query = join('#',@query).'#';
    $query =~ s/\*/[^#]+?/;

    #print "query: $query\n";

    my @current;

    push my @stack,[$this,0];

    while($#stack >= 0){
        my ($tag,$depth) = @{shift @stack};

        if($tag->{type} ne 'text'){
            splice @current, $depth;
            push @current, $tag->{value};
            my $current = join('#',@current).'#';

            #print "compare: $current =~ m/$query/\n";

            if($current =~ m/$query/){

                # print "($result): $current =~ m/$query/\n";

                if($result =~ m/\@(.*)/){
                    my $n = $1;
                    if($n){
                        for( @{ $tag->{children} } ){
                            if($_->{type} eq 'tag' && $_->{value} eq $n){
                                push @result,$_;
                            }
                        }
                    }else{
                        for( @{ $tag->{children} } ){
                            if($_->{type} eq 'tag'){
                                push @result,$_;
                            }
                        }
                    }
                }elsif($result =~ m/\%(.*)/){
                    # print "$result =~ m/\%(.*)/ \n";
                    my $n = $1;
                    # print "n is: $n\n";
                    if($n){
                        push @result,$tag->{attrib}{$n} if $tag->{attrib}{$n}; 
                    }else{
                        for(sort keys %{$tag->{attrib}}){
                            push @result,[$_,$tag->{attrib}{$_}];
                        }
                    }
                }elsif($result =~ m/\$(.*)/){
                    for( @{ $tag->{children} } ){
                        push @result,$_->{value} if $_->{type} eq 'text';
                    }
                }

            }else{
                for( @{ $tag->{children} } ){
                    push @stack,[$_,$depth+1];
                }
            }
        }
    }
    return wantarray ? @result : $result[0];
}

1;

